import argparse
from aergia._cli._command._chat import add_chat_subcommand
from aergia._cli._command._image import add_image_subcommand
from aergia._cli._command._models import add_models_subcommand
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def make_parser():
    parser = argparse.ArgumentParser(
        prog='ae[rgia]', formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    parser.add_argument(
        '--version', action='store_true', help='print the version'
    )
    parser.add_argument(
        '--debug', action='store_true', help='enable debug mode'
    )
    parser.add_argument(
        '--log-level', default=None,
        choices=['debug', 'info', 'warn', 'error', 'critical'],
        help='configure the log-level of the application'
    )

    subparsers = parser.add_subparsers(dest='subcommand')

    add_chat_subcommand(subparsers)
    # add_tui_subcommand(subparsers)
    # add_repl_subcommand(subparsers)
    # add_code_subcommand(subparsers)
    add_image_subcommand(subparsers)
    # add_execute_subcommand(subparsers)
    # add_session_subcommand(subparsers)
    # add_role_subcommand(subparsers)
    # add_info_subcommand(subparsers)
    add_models_subcommand(subparsers)

    return parser



import io
import os
import sys
import httpx
import datetime as dt
from openai import OpenAI
from PIL import Image as PillowImage
from rich.table import Table
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from aergia._model._storage import application_db, save, load_list, load_blob
from aergia._model._data import Image
from aergia._cli._command._utilities import ExitCode


def image(args, stdout, stderr):
    def download(url, file):
        resp = httpx.get(url)
        file.write(resp.content)

    client = OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    if not args.prompt:
        prompt = sys.stdin.read()
    else:
        prompt = args.prompt if isinstance(args.prompt, str) else " ".join(args.prompt)
    response = client.images.generate(prompt=prompt, model=args.model)
    data = response.data[0]
    image_url = data.url
    revised_prompt = data.revised_prompt
    buffer = io.BytesIO()
    download(image_url, buffer)
    created = dt.datetime.fromtimestamp(response.created)
    img = Image(
        id=None,
        name=args.name,
        model=args.model,
        prompt=prompt,
        created=created,
        revised_prompt=revised_prompt,
        blob=buffer.getvalue()
    )
    db = application_db()
    save(img, db)

    pimg = PillowImage.open(buffer)
    pimg.show(args.name)

    return ExitCode.Success


def list_images(args, stdout, stderr):
    table = Table(title='Generated Images')
    table.add_column("Id", justify='right')
    table.add_column("Name", justify='left', style='green')
    table.add_column("Model", justify='left', style='cyan')
    table.add_column("Created", justify='left', style='yellow')
    table.add_column("Prompt", justify='left', style='white')
    table.add_column("Revised-Prompt", justify='left', style='magenta')

    db = application_db()
    limit = None if args.all else args.limit
    images = load_list(Image, db, limit=limit, offset=args.offset)
    for img in images:
        table.add_row(
            f"{img.id}",
            img.name,
            img.model,
            f"{img.created}",
            img.prompt,
            img.revised_prompt
        )

    stdout.print(table)

    return ExitCode.Success


def show_image(args, stdout, stderr):
    db = application_db()
    blob = load_blob(Image, args.id, db)
    pimg = PillowImage.open(blob)
    pimg.show()
    return ExitCode.Success


def add_image_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'image',
        help='manage and generate images',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    sub_subparsers = subcommand.add_subparsers(
        dest='session_subcommand', help='Sessions commands'
    )

    generate_subcommand = sub_subparsers.add_parser(
        'generate',
        help='generate images',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )

    generate_subcommand.add_argument(
        '-n', '--name', type=str, default='unnamed',
        help='name or title which will be used for the image'
    )
    generate_subcommand.add_argument(
        '-m', '--model',
        default='dall-e-2',
        choices=['dall-e-2', 'dall-e-3'],
        help='model to use for generating the image'
    )
    generate_subcommand.add_argument(
        'prompt', type=str, nargs="*",
        metavar='PROMPT',
        help='prompt which shall be used for generating the image, if None stdin will be read.'
    )
    generate_subcommand.set_defaults(func=image)

    list_subcommand = sub_subparsers.add_parser(
        'list',
        help='list previously generated images',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    list_subcommand.add_argument(
        '-a', '--all', action='store_true', default=False, help='list all images'
    )
    list_subcommand.add_argument(
        '-l', '--limit', type=int, default=10, help='amount of images to list'
    )
    list_subcommand.add_argument(
        '-o', '--offset', type=int, default=0, help='offset for listing images'
    )
    list_subcommand.set_defaults(func=list_images)

    show_subcommand = sub_subparsers.add_parser(
        'show',
        help='show an image',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    show_subcommand.add_argument(
        'id', type=int, help='image to show'
    )
    show_subcommand.set_defaults(func=show_image)
    return subcommand



from __future__ import annotations

import datetime
import io
import sqlite3
from pathlib import Path
from inspect import cleandoc

from aergia._model._data import Session, Message, Image

TABLES = [Image, Session, Message]


def data_directory(name: str) -> Path:
    return Path.home() / '.local' / 'share' / name


def store(name, directory):
    directory.mkdir(exist_ok=True, parents=True)
    db = directory / f"{name}.sqlite"
    with sqlite3.connect(db):
        pass
    return db


def initialize(db, tables=None):
    tables = tables or []
    with sqlite3.connect(db) as con:
        with con as transaction:
            for table in tables:
                transaction.execute(table.ddl)


def application_db():
    name = 'aergia'
    directory = data_directory(name)
    db = store(name, directory)
    initialize(db, TABLES)
    return db


def save(image: Image, db=None):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc("""
            INSERT INTO images (name, model, prompt, revised_prompt, created, blob)
            VALUES (?, ?, ?, ?, ?, ?);
            """)
            result = transaction.execute(
                stmt,
                (
                    image.name,
                    image.model,
                    image.prompt,
                    image.revised_prompt or "[No Revised Prompt]",
                    image.created.timestamp(),
                    image.blob,
                )
            )
        image.id = result.lastrowid
        return image


def load(type, name, db=None) -> Image:
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc("""
            SELECT id, name, model, prompt, revised_prompt, created, blob FROM images
            WHERE name = ?;
            """)
            result = transaction.execute(stmt, (name,))
            row = result.fetchone()
            names = Image.model_json_schema()['properties'].keys()
            kwargs = dict(zip(names, row))
            kwargs['created'] = datetime.datetime.fromtimestamp(kwargs['created'])
            return Image(**kwargs)


def load_list(type, db=None, limit=10, offset=0):
    with sqlite3.connect(db) as con:
        if limit is None:
            stmt = "SELECT id, name, model, prompt, revised_prompt, created, blob FROM images;"
            result = con.execute(stmt)
        else:
            stmt = cleandoc("""
                SELECT id, name, model, prompt, revised_prompt, created, blob FROM images
                ORDER BY id
                LIMIT ? OFFSET ?;
            """)
            result = con.execute(stmt, (limit, offset))
        names = Image.model_json_schema()['properties'].keys()
        rows = result.fetchall()
        images = []
        for row in rows:
            kwargs = dict(zip(names, row))
            kwargs['created'] = datetime.datetime.fromtimestamp(kwargs['created'])
            images.append(Image(**kwargs))
        return images


def load_blob(type, id, db=None):
    with sqlite3.connect(db) as con:
        stmt = "SELECT blob FROM images WHERE id = ?;"
        result = con.execute(stmt, (id,))
        row = result.fetchone()
        return io.BytesIO(row[0])



import argparse
from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_code_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'code',
        help='Generate code',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Code prompt'
    )
    subcommand.add_argument(
        '-r', '--role', type=str, help='Role to be used in addition'
    )
    subcommand.set_defaults(func=default)
    return subcommand



import os
import asyncio
import argparse
from openai import AsyncOpenAI
from rich.markdown import Markdown
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from aergia._cli._command._utilities import TextBuffer, ExitCode


def chat(args, stdout, stderr):
    async def achat(client, model, message):
        stream = await client.chat.completions.create(
            model=model,
            messages=[{"role": "user", "content": message}],
            stream=True,
        )
        with stdout.status("Processing request ...", spinner='aesthetic', spinner_style='cyan') as status:
            with TextBuffer() as buffer:
                async for chunk in stream:
                    text = chunk.choices[0].delta.content or ""
                    buffer.append(text)
                content = buffer.content

        stdout.print(
            Markdown(content)
        )
        return ExitCode.Success

    client = AsyncOpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    msg = " ".join(args.text)
    context = "" if not args.context else args.context.read()
    msg = msg + context
    return asyncio.run(achat(client=client, model=args.model, message=msg))


def add_chat_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'chat',
        help='chat with the ai system',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Text input'
    )
    subcommand.add_argument(
        '-r', '--role', type=str, help='Choose a role'
    )
    subcommand.add_argument(
        '-s', '--session', nargs='?', const=True, help='Create or reuse a session'
    )
    subcommand.add_argument(
        '-c', '--context', type=argparse.FileType('r'), help='File directory or stdin'
    )
    subcommand.add_argument(
        '-m', '--model',
        default='gpt-4o',
        choices=['gpt-4o', 'gpt-4', 'gpt-4-turbo', 'gpt-3.5-turbo'],
        help='Select a model'
    )
    subcommand.add_argument(
        '--stream', dest='stream', action='store_true', default=True, help='Enable streaming'
    )
    subcommand.set_defaults(func=chat)
    return subcommand



from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_session_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'session',
        help='Manage sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)

    sub_subparsers = subcommand.add_subparsers(
        dest='session_subcommand', help='Sessions commands'
    )

    search_command = sub_subparsers.add_parser(
        'search',
        help='Search for text in sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    search_command.add_argument('text', type=str, help='Text to search')

    list_command = sub_subparsers.add_parser(
        'list',
        help='List all sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )

    show_command = sub_subparsers.add_parser(
        'show',
        help='Show contents of a session',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )

    alias_command = sub_subparsers.add_parser(
        'alias',
        help='Alias for session command',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    return subcommand



from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_role_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'role',
        help='Manage roles',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)

    sub_subparsers = subcommand.add_subparsers(
        dest='role_subcommand', help='Roles commands'
    )

    list_command = sub_subparsers.add_parser(
        'list', help='List all roles'
    )

    show_command = sub_subparsers.add_parser(
        'show', help='Show a role'
    )
    show_command.add_argument(
        'name', type=str, help='Role name'
    )

    create_command = sub_subparsers.add_parser(
        'create', help='Create a role'
    )
    create_command.add_argument(
        'name', type=str, help='Role name'
    )
    create_command.add_argument(
        'text', type=str, nargs='?', help='Role description'
    )

    update_command = sub_subparsers.add_parser(
        'update', help='Update a role'
    )
    update_command.add_argument(
        'name', type=str, help='Role name'
    )
    update_command.add_argument(
        'text', type=str, nargs='?', help='New role description'
    )

    delete_command = sub_subparsers.add_parser(
        'delete', help='Delete a role'
    )
    delete_command.add_argument(
        'name', type=str, help='Role name'
    )

    alias_command = sub_subparsers.add_parser(
        'alias', help='Create an alias for a role'
    )
    alias_command.add_argument(
        'role', type=str, help='Role name'
    )
    alias_command.add_argument(
        'alias', type=str, help='Alias name'
    )
    return subcommand



from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_tui_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'tui',
        help='start textul based tui client',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)
    return subcommand



from __future__ import annotations

import datetime as dt
from inspect import cleandoc

from pydantic import BaseModel


class Session:
    _TABLE_NAME = 'session'

    id: int
    name: str
    model: str
    temperature: float

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE IF NOT EXISTS {cls._TABLE_NAME} (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            model TEXT NOT NULL,
            temperature REAL NOT NULL
        );
        """)


class Message:
    _TABLE_NAME = 'messages'

    id: int
    role: str
    content: str
    session_id: int

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE IF NOT EXISTS {cls._TABLE_NAME} (
            id INTEGER PRIMARY KEY,
            role TEXT NOT NULL,
            content TEXT NOT NULL,
            session_id INTEGER,
            FOREIGN KEY (session_id) REFERENCES sessions(id)
        );
        """)


class Image(BaseModel):
    _TABLE_NAME = 'images'
    id: int | None
    name: str
    model: str
    prompt: str
    revised_prompt: str | None
    created: dt.datetime
    blob: bytes

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE  IF NOT EXISTS {cls._TABLE_NAME.default} (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL,
            model TEXT NOT NULL,
            prompt TEXT NOT NULL,
            revised_prompt TEXT NOT NULL,
            created INTEGER NOT NULL, 
            blob BLOB NOT NULL
        );
        """)
